/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/*          EXIFPROBE - TIFF/JPEG/EXIF image file probe               */
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/* Copyright (C) 2002 by Duane H. Hesser. All rights reserved.        */
/*                                                                    */
/* See the file LICENSE.EXIFPROBE for terms of use.                   */
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */

#ifndef lint
static char *ModuleId = "@(#) $Id: interpret.c,v 1.14 2004/12/04 16:00:28 alex Exp $";
#endif

#include <stdio.h>
#include "defs.h"
#include "extern.h"
#include "misc.h"
#include "tags.h"

/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/* This file contains the functions which interpret and print         */
/* "meaningful" descriptions for values found in TIFF and EXIF ifds.  */
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */



static void print_aperture(FILE *,unsigned short,unsigned long);
static void print_brightness(FILE *,unsigned short,unsigned long);
static void print_shutterspeed(FILE *,unsigned short,unsigned long);

static void print_capturetype(unsigned long);
static void print_colorspace(unsigned long);
static void print_componentsconfig(unsigned long);
static void print_compression(unsigned long);
static void print_contrast(unsigned long);
static void print_custom(unsigned long);
static void print_epfpresunit(unsigned long);
static void print_exposuremode(unsigned long);
static void print_exposureprogram(unsigned long);
static void print_filesource(unsigned long source);
static void print_flash(unsigned long flash);
static void print_gaincontrol(unsigned long);
static void print_jpegproc(unsigned long);
static void print_lightsource(unsigned long);
static void print_meteringmode(unsigned long);
static void print_newsubfiletype(unsigned long);
static void print_oldsubfiletype(unsigned long);
static void print_orientation(unsigned long);
static void print_photometric(unsigned long);
static void print_planarconfig(unsigned long);
static void print_resunit(unsigned long);
static void print_saturation(unsigned long);
static void print_scenetype(unsigned long type);
static void print_sensingmethod(unsigned long);
static void print_sharpness(unsigned long);
static void print_subjectdistancerange(unsigned long);
static void print_whitebalance(unsigned long);
static void print_ycbcrpositioning(unsigned long);

static void print_35mmfocallength();

void
interpret_value(struct ifd_entry *entry_ptr)
{
    switch(entry_ptr->tag)
    {
        case TIFFTAG_Compression:
            print_compression(entry_ptr->value);
            break;
        case TIFFTAG_PhotometricInterpretation:
            print_photometric(entry_ptr->value);
            break;
        case TIFFTAG_Orientation:
            print_orientation(entry_ptr->value);
            break;
        case TIFFTAG_ResolutionUnit:
        case EXIFTAG_FocalPlaneResolutionUnit:
            print_resunit(entry_ptr->value);
            break;
        case TIFFEPTAG_FocalPlaneResolutionUnit:
            print_epfpresunit(entry_ptr->value);
            break;
        case EXIFTAG_ColorSpace:
            print_colorspace(entry_ptr->value);
            break;
        case EXIFTAG_MeteringMode:
            print_meteringmode(entry_ptr->value);
            break;
        case EXIFTAG_SensingMethod:
            print_sensingmethod(entry_ptr->value);
            break;
        case EXIFTAG_ExposureProgram:
            print_exposureprogram(entry_ptr->value);
            break;
        case EXIFTAG_ComponentsConfiguration:
            print_componentsconfig(entry_ptr->value);
            break;
        case TIFFTAG_SubFileType:
            print_oldsubfiletype(entry_ptr->value);
            break;
        case TIFFTAG_NewSubFileType:
            print_newsubfiletype(entry_ptr->value);
            break;
        case TIFFTAG_PlanarConfiguration:
            print_planarconfig(entry_ptr->value);
            break;
        case TIFFTAG_JpegProc:
            print_jpegproc(entry_ptr->value);
            break;
        case TIFFTAG_YcbCrPositioning:
            print_ycbcrpositioning(entry_ptr->value);
            break;
        case EXIFTAG_LightSource:
            print_lightsource(entry_ptr->value);
            break;
        case EXIFTAG_Flash:
            print_flash(entry_ptr->value);
            break;
        case  EXIFTAG_FileSource:
            print_filesource(entry_ptr->value);
            break;
        case  EXIFTAG_SceneType:
            print_scenetype(entry_ptr->value);
            break;
        case EXIFTAG_ExifCustomRendered:
            print_custom(entry_ptr->value);
            break;
        case EXIFTAG_ExifExposureMode:
            print_exposuremode(entry_ptr->value);
            break;
        case EXIFTAG_ExifWhiteBalance:
            print_whitebalance(entry_ptr->value);
            break;
        case EXIFTAG_ExifDigitalZoomRatio:
            /* Nothing to do                                          */
            break;
        case EXIFTAG_ExifFocalLengthIn35mmFilm:
            print_35mmfocallength();
            break;
        case EXIFTAG_ExifSceneCaptureType:
            print_capturetype(entry_ptr->value);
            break;
        case EXIFTAG_ExifGainControl:
            print_gaincontrol(entry_ptr->value);
            break;
        case EXIFTAG_ExifContrast:
            print_contrast(entry_ptr->value);
            break;
        case EXIFTAG_ExifSaturation:
            print_saturation(entry_ptr->value);
            break;
        case EXIFTAG_ExifSharpness:
            print_sharpness(entry_ptr->value);
            break;
        case EXIFTAG_ExifSubjectDistanceRange:
            print_subjectdistancerange(entry_ptr->value);
            break;
        case EXIFTAG_ExifImageUniqueId:
            /* Nothing to do                                          */
            break;
        default:
            break;
    }
}

void
print_compression(unsigned long compvalue)
{
    char *compstr;

    if((compstr = comp_format((int)compvalue)))
    {
        print_startvalue();
        printf("%s",compstr);
        print_endvalue();
    }
}


void
print_photometric(unsigned long metric)
{
    print_startvalue();
    switch(metric)
    {
        case 0: printf("white is 0"); break;
        case 1: printf("black is 0"); break;
        case 2: printf("RGB"); break;
        case 3: printf("Palette"); break;
        case 4: printf("transparency mask"); break;
        case 5: printf("CMYK"); break;
        case 6: printf("YCbCr"); break;
        case 8: printf("CIELAB"); break;
        case 10: printf("ITULAB"); break;
        case 32803: printf("CFA"); break;    /* TIFFEP */
        case 32844: printf("CIE Log2(L)"); break;
        case 32845: printf("CIE Log2(L) (u',v')"); break;
        default: printf("undefined"); break;
    }
    print_endvalue();
}

void
print_orientation(unsigned long value)
{
    print_startvalue();
    switch(value)
    {
        case 1: printf("0,0 is top left"); break;
        case 2: printf("0,0 is top right"); break;
        case 3: printf("0,0 is bottom right"); break;
        case 4: printf("0,0 is bottom left"); break;
        case 5: printf("0,0 is left top"); break;
        case 6: printf("0,0 is right top"); break;
        case 7: printf("0,0 is right bottom"); break;
        case 8: printf("0,0 is left bottom"); break;
        default: printf("undefined"); break;
    }
    print_endvalue();
}

void
print_resunit(unsigned long unit)
{
    print_startvalue();
    switch(unit)
    {
        case 1: printf("no units"); break;
        case 2: printf("pixels per inch"); break;
        case 3: printf("pixels per cm"); break;
        default: printf("undefined"); break;
    }
    print_endvalue();
}

void
print_epfpresunit(unsigned long unit)
{
    print_startvalue();
    switch(unit)
    {
        case 1: printf("no units"); break;
        case 2: printf("per meter"); break;
        case 3: printf("per cm"); break;
        case 4: printf("per mm"); break;
        case 5: printf("per um"); break;
        default: printf("undefined"); break;
    }
    print_endvalue();
}

void
print_colorspace(unsigned long space)
{
    print_startvalue();
    if(space == 1)
        printf("sRGB");
    else if(space == 0xffff)
        printf("uncalibrated");
    else
        printf("reserved");
    print_endvalue();
}

void
print_meteringmode(unsigned long mode)
{
    print_startvalue();
    switch(mode)
    {
        case 0: printf("unknown"); break;
        case 1: printf("Average"); break;
        case 2: printf("Center Weighted Average"); break;
        case 3: printf("Spot"); break;
        case 4: printf("Multi Spot"); break;
        case 5: printf("Pattern"); break;
        case 6: printf("Partial"); break;
        default: if(mode < 255) printf("reserved");
            else printf("other"); break;
    }
    print_endvalue();
}


void
print_sensingmethod(unsigned long method)
{
    print_startvalue();
    switch(method)
    {
        case 1: printf("Not defined"); break;
        case 2: printf("One-chip color area sensor"); break;
        case 3: printf("Two-chip color area sensor"); break;
        case 4: printf("Three-chip color area sensor"); break;
        case 5: printf("Color sequential area sensor"); break;
        case 7: printf("Trilinear sensor"); break;
        case 8: printf("Color sequential linear sensor"); break;
        default: printf("reserved"); break;
    }
    print_endvalue();
}

void
print_exposureprogram(unsigned long program)
{
    print_startvalue();
    switch(program)
    {
        case 0: printf("Not defined"); break;
        case 1: printf("Manual"); break;
        case 2: printf("Normal"); break;
        case 3: printf("Aperture Priority"); break;
        case 4: printf("Shutter Priority"); break;
        case 5: printf("Creative Program"); break;
        case 6: printf("Action Program"); break;
        case 7: printf("Portrait Mode"); break;
        case 8: printf("Landscape Mode"); break;
        default:
            if(program < 255)
                printf("reserved");
            break;
    }
    print_endvalue();
}

void
print_componentsconfig(unsigned long components)
{
    char *comp = (char *)&components;
    int i;

    print_startvalue();
    for(i = 0; i < 4; ++i)
    {
        switch(*comp++)
        {
        case 1: putchar('Y'); break;
        case 2: putchar('C'); putchar('b'); break;
        case 3: putchar('C'); putchar('r'); break;
        case 4: putchar('R'); break;
        case 5: putchar('G'); break;
        case 6: putchar('B'); break;
        case 0: break;
        default:
            putchar('.');
            break;
        }
    }
    print_endvalue();
}


void
print_newsubfiletype(unsigned long type)
{
    print_startvalue();
    switch(type)
    {
    case 0: printf("primary"); break;
    case 1: printf("thumbnail"); break;
    case 2: printf("page"); break;
    case 4: printf("mask"); break;
    default: printf("unknown"); break;
    }
    print_endvalue();
}


void
print_oldsubfiletype(unsigned long type)
{
    print_startvalue();
    if(type == 1)
        printf("primary");
    else if(type == 2)
        printf("thumbnail");
    else
        printf("unknown");
    print_endvalue();
}

void
print_planarconfig(unsigned long config)
{
    print_startvalue();
    if(config == 1)
        printf("chunky/contig");
    else if(config == 2)
        printf("planar/separate");
    else
        printf("unknown");
    print_endvalue();
}

void
print_jpegproc(unsigned long process)
{
    print_startvalue();
    if(process == 1)
        printf("baseline");
    else if(process == 14)
        printf("lossless");
    else
        printf("unknown");
    print_endvalue();
}

void
print_ycbcrpositioning(unsigned long position)
{
    print_startvalue();
    if(position == 1)
        printf("centered");
    else if(position == 2)
        printf("co-sited");
    else
        printf("unknown");
    print_endvalue();
}

void
print_lightsource(unsigned long source)
{
    print_startvalue();
    switch(source)
    {
        case 0: printf("unknown"); break;
        case 1: printf("Daylight"); break;
        case 2: printf("Fluorescent"); break;
        case 3: printf("Tungsten"); break;
        case 4: printf("Flash"); break;
        case 9: printf("Fine weather"); break;
        case 10: printf("Cloudy weather"); break;
        case 11: printf("Shade"); break;
        case 12: printf("Daylight fluorescent"); break;
        case 13: printf("Day white fluorescent"); break;
        case 14: printf("Cool white fluorescent"); break;
        case 15: printf("Fluorescent"); break;
        case 17: printf("Standard light A"); break;
        case 18: printf("Standard light B"); break;
        case 19: printf("Standard light C"); break;
        case 20: printf("D65"); break;
        case 21: printf("D55"); break;
        case 22: printf("D75"); break;
        case 23: printf("D50"); break;
        case 24: printf("ISO studoio tungsten"); break;
        case 25: printf("Other light source"); break;
        default:
                printf("reserved");
            break;
    }
    print_endvalue();
}

void
print_flash(unsigned long flash)
{
    unsigned long bits;

    print_startvalue();
    bits = flash & 0x1;
    if(bits != 0)
        printf("flash fired");
    else
        printf("flash did not fire");
    bits = (flash & 0x6) >> 1;
    switch(bits)    /* return detection                               */
    {
        case 2: printf(" - return not detected"); break;
        case 3: printf(" - return detected"); break;
        default: break;
    }
    bits = (flash & 0x18) >> 3; /* flash mode                         */
    switch(bits)
    {
        case 1: printf(" - compulsory"); break;
        case 2: printf(" - suppressed"); break;
        case 3: printf(" - auto"); break;
        default: break;
    }
    bits = (flash & 0x20) >> 5; /* flash function                     */
    if(bits != 0)
        printf(" - no flash function present");
    bits = (flash & 0x40) >> 6; /* red-eye                            */
    if(bits != 0)
        printf(" - no red-eye function present");
    print_endvalue();
}

void
print_filesource(unsigned long source)
{
    print_startvalue();
    if(source == 3)
        printf("DSC");
    else
        printf("not DSC");
    print_endvalue();
}

void
print_scenetype(unsigned long type)
{
    print_startvalue();
    if(type == 1)
        printf("direct photo");
    else
        printf("not direct photo");
    print_endvalue();
}

void
print_custom(unsigned long rendered)
{
    print_startvalue();
    if(rendered == 0)
        printf("Normal");
    else if(rendered == 1)
        printf("Custom");
    else
        printf("reserved");
    print_endvalue();
}

void
print_exposuremode(unsigned long mode)
{
    print_startvalue();
    switch(mode)
    {
        case 0: printf("Auto"); break;
        case 1: printf("Manual"); break;
        case 2: printf("Auto Bracket"); break;
        default: printf("???"); break;
    }
    print_endvalue();
}

void
print_whitebalance(unsigned long mode)
{
    print_startvalue();
    switch(mode)
    {
        case 0: printf("Auto"); break;
        case 1: printf("Manual"); break;
        default: printf("???"); break;
    }
    print_endvalue();
}

void
print_35mmfocallength()
{
    /* %%%### */
    printf("mm");
}

void
print_capturetype(unsigned long type)
{
    print_startvalue();
    switch(type)
    {
        case 0: printf("Standard"); break;
        case 1: printf("Landscape"); break;
        case 2: printf("Portrait"); break;
        case 3: printf("Night Scene"); break;
        default: printf("reserved"); break;
    }
    print_endvalue();
}

void
print_gaincontrol(unsigned long type)
{ 
    print_startvalue();
    switch(type)
    {
        case 0: printf("None"); break;
        case 1: printf("Low gain up"); break;
        case 2: printf("High gain up"); break;
        case 3: printf("Low gain down"); break;
        case 4: printf("Hign gain down"); break;
        default: printf("reserved"); break;
    }
    print_endvalue();
}

void
print_contrast(unsigned long contrast)
{
    print_startvalue();
    switch(contrast)
    {
        case 0: printf("Normal"); break;
        case 1: printf("Soft"); break;
        case 2: printf("Hard"); break;
        default: printf("???"); break;
    }
    print_endvalue();
}

void
print_saturation(unsigned long saturation)
{
    print_startvalue();
    switch(saturation)
    {
        case 0: printf("Normal"); break;
        case 1: printf("Low"); break;
        case 2: printf("High"); break;
        default: printf("???"); break;
    }
    print_endvalue();
}

void
print_sharpness(unsigned long sharpness)
{
    print_startvalue();
    switch(sharpness)
    {
        case 0: printf("Normal"); break;
        case 1: printf("Soft"); break;
        case 2: printf("Hard"); break;
        default: printf("???"); break;
    }
    print_endvalue();
}

void
print_subjectdistancerange(unsigned long range)
{
    print_startvalue();
    switch(range)
    {
        case 0: printf("???"); break;
        case 1: printf("Macro"); break;
        case 2: printf("Close view"); break;
        case 3: printf("Distant view"); break;
        default: printf("???"); break;
    }
    print_endvalue();
}

void
interpret_rational(FILE *inptr,unsigned short byteorder,unsigned short tag,unsigned long offset)
{
    if(inptr)
    {
        switch(tag)
        {
            case EXIFTAG_ShutterSpeedValue:
                printf(" APEX");    /* ###%%% wrong place to put this.            */
                print_shutterspeed(inptr,byteorder,offset);
                break;
            case EXIFTAG_ApertureValue:
            case EXIFTAG_MaxApertureValue:
                printf(" APEX");    /* ###%%% wrong place to put this.            */
                print_aperture(inptr,byteorder,offset);
                break;
            case EXIFTAG_BrightnessValue:
                printf(" APEX");    /* ###%%% wrong place to put this.            */
                print_brightness(inptr,byteorder,offset);
                break;
            case EXIFTAG_FocalLength:
                printf(" mm");
                break;
            case EXIFTAG_SubjectDistance:
                printf(" meters");
                break;
            case EXIFTAG_ExposureBiasValue:
                printf(" APEX");    /* ###%%% wrong place to put this.            */
                break;
            default:
                break;
        }
    }
}

#include <math.h>

void
print_aperture(FILE *inptr,unsigned short byteorder,unsigned long offset)
{
    double aperture_value,fnumber;
    unsigned long num,denom;

    num = read_ulong(inptr,byteorder,offset);
    denom = read_ulong(inptr,byteorder,HERE);
    aperture_value = (double)((double)num/(double)denom);
    fnumber = pow(2.0,aperture_value/2.0);
    print_startvalue();
    printf("f%.2g",fnumber);
    print_endvalue();
}

void
print_shutterspeed(FILE *inptr,unsigned short byteorder,unsigned long offset)
{
    double speed_value,exptime;
    long num,denom;

    num = read_ulong(inptr,byteorder,offset);
    denom = read_ulong(inptr,byteorder,HERE);
    speed_value = (double)((double)num/(double)denom);
    exptime = pow(2.0,-speed_value);
    print_startvalue();
    printf("%g sec",exptime);
    print_endvalue();
}

void
print_brightness(FILE *inptr,unsigned short byteorder,unsigned long offset)
{
    double brightness_value,brightness;
    long num,denom;

    num = read_ulong(inptr,byteorder,offset);
    denom = read_ulong(inptr,byteorder,HERE);
    brightness_value = (double)((double)num/(double)denom);
    brightness = pow(2.0,brightness_value);
    print_startvalue();
    printf("%.3g foot lambert",brightness);
    print_endvalue();
}
