/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/*          EXIFPROBE - TIFF/JPEG/EXIF image file probe               */
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/* Copyright (C) 2002 by Duane H. Hesser. All rights reserved.        */
/*                                                                    */
/* See the file LICENSE.EXIFPROBE for terms of use.                   */
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */

#ifndef lint
static char *ModuleId = "@(#) $Id: maker_leica.c,v 1.5 2003/04/26 14:15:17 alex Exp $";
#endif

/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/* Leica camera maker-specific routines                               */
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/* This information is from a page by Johannes Tschebisch at:         */
/* http://www.jojotsch.de/downloads/jojothumb/beispiele/              */
/*                             html_exif/bilder/leica-digilux4.3.html */
/*                                                                    */
/* Note that the tags are identical to the fujifilm tags, the format  */
/* is the same wierd format as the fuji, and the note even begins     */
/* with the string "FUJIFILM".                                        */
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */

#include <stdio.h>
#include <string.h>
#include "defs.h"
#include "extern.h"
#include "misc.h"
#include "maker.h"

/* Find the identifying number assigned to known Leica camera         */
/* models. This number is used to dispatch print and interpret        */
/* routines approopriate to the current image.                        */

int
leica_model_number(char *model,char *software)
{
    int number = GENERIC;

    if(model)
    {
        switch(*model)
        {
            case 'd':
                if(strncmp(model,"digilux 4.3",10) == 0)
                    number = LEICA_DIGILUX43;
                break;
            default:
                /* until told otherwise...                                */
                number = LEICA_DIGILUX43;
                break;
        }
    }
    return(number);
}


/* Dispatch a print routine for direct values in Leica cameras,       */
/* based upon model                                                   */

void
print_leica_makervalue(struct ifd_entry *entry_ptr,int make,int model,
                    char *prefix)
{

    if(entry_ptr && (PRINT_VALUE))
    {
        switch(model)
        {
            case LEICA_DIGILUX43:
            case GENERIC:
            default:
                print_leica1_makervalue(entry_ptr,make,model,prefix);
                leica1_interpret_value(entry_ptr);
                break;
        }
    }
}

/* Model-specific print routine for Leica cameras. This routine is    */
/* responsible for picking off any direct entry tags which are        */
/* peculiar and will not be handled properly by print_value()         */
/* (usually UNDEFINED values which fit in the 4-byte entry value). If */
/* there are no such entries, this function simply calls              */
/* print_value().                                                     */

void
print_leica1_makervalue(struct ifd_entry *entry_ptr,int make,int model,
                    char *prefix)
{
    if(entry_ptr && (PRINT_VALUE))
    {
        switch(entry_ptr->tag) 
        {
            default:
                print_value(entry_ptr,PREFIX);
                break;
        }
    }
}


/* Dispatch a routine to decode and print offset values for Leica     */
/* cameras.                                                           */

void
print_leica_offset_makervalue(FILE *inptr,unsigned short byteorder,
    struct ifd_entry *entry_ptr,unsigned long offset_ref,char *prefix,
    int indent,int addrwidth,int make,int model,int at_offset)
{

    if(entry_ptr && (PRINT_VALUE))
    {
        switch(model)
        {
            /* currently only one routine available                       */
            default:
                leica1_offset_makervalue(inptr,byteorder,entry_ptr,
                                            offset_ref,prefix,indent,
                                            addrwidth,make,model,at_offset);
                break;
        }
    }
}

/* Model-specific routine to print UNDEFINED values found at offsets  */
/* in Leica makernotes.                                               */

void
leica1_offset_makervalue(FILE *inptr,unsigned short byteorder,
    struct ifd_entry *entry_ptr,unsigned long offset_ref,char *prefix,
    int indent,int addrwidth,int make,int model,int at_offset)
{
    if(entry_ptr)
    {
        switch(entry_ptr->tag)
        {
            default:
                print_generic_offset_makervalue(inptr,byteorder,entry_ptr,
                                        offset_ref,prefix,indent,
                                        addrwidth,make,model,at_offset);
                break;
        }
    }
}


/* Leica-specific tagnames for makernotes.                            */

char *
maker_leica_tagname(unsigned short tag,int model)
{
    char *tagname = (char *)0;

    /* Check specific models first                                    */
    switch(model)
    {
        /* Add special-case models here                               */
        default:
            break;
    }

    /* If no model-specific tag is found, use a "generic" Leica       */
    /* tag                                                            */
    if(tagname == NULL)
    {
        switch(tag)
        {
            case 0x0000: tagname = "Version"; break;
            case 0x1000: tagname = "Quality"; break;
            case 0x1001: tagname = "Sharpness"; break;
            case 0x1002: tagname = "WhiteBalance"; break;
            case 0x1003: tagname = "Color"; break;
            case 0x1004: tagname = "Tone"; break;
            case 0x1010: tagname = "FlashMode"; break;
            case 0x1011: tagname = "FlashStrength"; break;
            case 0x1020: tagname = "Macro"; break;
            case 0x1021: tagname = "FocusMode"; break;
            case 0x1030: tagname = "SlowSync"; break;
            case 0x1031: tagname = "PictureMode"; break;
            case 0x1100: tagname = "ContTake-Bracket"; break;
            case 0x1300: tagname = "BlurWarning"; break;
            case 0x1301: tagname = "FocusWarning"; break;
            case 0x1302: tagname = "AEWarning"; break;
            default: break;
        }
    }
    return(tagname);
}


/* Report the "meaning" of tag values found at offsets in a Leica     */
/* MakerNote IFD entry (not at an offset).                            */

void
leica1_interpret_value(struct ifd_entry *entry_ptr)
{
    switch(entry_ptr->tag)
    {
        case 0x1001:    /* Sharpness */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 1: printf("Soft"); break;
                case 2: printf("Soft"); break;
                case 3: printf("Normal"); break;
                case 4: printf("Hard"); break;
                case 5: printf("Hard"); break;
                default: printred("undefined"); break;
            }
            print_endvalue();
            break;
        case 0x1002:    /* White Balance */
            print_startvalue();
            /* I think we're looking at a bit mask here, but I have   */
            /* no way to check; this is the way Tachibanaya           */
            /* characterizes it.                                      */
            switch(entry_ptr->value)
            {
                case 0: printf("Auto"); break;
                case 256: printf("Daylight"); break;
                case 512: printf("Cloudy"); break;
                case 768:   /* ???? */ printf("DaylightColor-Fluorescent"); break;
                case 769:   /* ???? */ printf("DaywhiteColor-Fluorescent"); break;
                case 770:   /* ???? */ printf("White-Fluorescent"); break;
                case 1024: printf("Incandescent"); break;
                case 3840: printf("Custom"); break;
                default: printred("undefined"); break;
            }
            print_endvalue();
            break;
        case 0x1003:    /* Color Saturation */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("Standard"); break;
                case 256: printf("High"); break;
                case 512: printf("Low"); break;
                default: printred("undefined"); break;
            }
            print_endvalue();
            break;
        case 0x1004:    /* Tone */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("Standard"); break;
                case 256: printf("High"); break;
                case 512: printf("Low"); break;
                default: printred("undefined"); break;
            }
            print_endvalue();
            break;
        case 0x1010:    /* Flash Mode */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("Auto"); break;
                case 1: printf("On"); break;
                case 2: printf("Off"); break;
                case 3: printf("Redeye"); break;
                default: printred("undefined"); break;
            }
            print_endvalue();
            break;
        case 0x1020:    /* Macro Mode */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("Off"); break;
                case 1: printf("On"); break;
                default: printred("undefined"); break;
            }
            print_endvalue();
            break;
        case 0x1021:    /* Focusing Mode */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("Auto"); break;
                case 1: printf("Manual"); break;
                default: printred("undefined"); break;
            }
            print_endvalue();
            break;
        case 0x1030:    /* Slow Sync */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("Off"); break;
                case 1: printf("On"); break;
                default: printred("undefined"); break;
            }
            print_endvalue();
            break;
        case 0x1031:    /* Picture Mode */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("Auto"); break;
                case 1: printf("Portrait"); break;
                case 2: printf("Landscape"); break;
                case 4: printf("Sports"); break;
                case 5: printf("Night"); break;
                case 6: printf("Program Normal"); break;
                case 256: printf("Aperture Priority"); break;
                case 512: printf("Shutter Priority"); break;
                case 768: printf("Manual"); break;
                default: printred("undefined");
            }
            print_endvalue();
            break;
        case 0x1100:    /* Continuous Take/Bracket */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("Off"); break;
                case 1: printf("On"); break;
                default: printred("undefined"); break;
            }
            print_endvalue();
            break;
        case 0x1300:    /* Blur Warning */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("Ok"); break;
                case 1: printf("Blurred"); break;
                default: printred("undefined"); break;
            }
            print_endvalue();
            break;
        case 0x1301:    /* Focus Warning */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("Ok"); break;
                case 1: printf("Out of focus"); break;
                default: printred("undefined"); break;
            }
            print_endvalue();
            break;
        case 0x1302:    /* AE Warning */
            print_startvalue();
            switch(entry_ptr->value)
            {
                case 0: printf("Ok"); break;
                case 1: printf("Overexposed"); break;
                default: printred("undefined"); break;
            }
            print_endvalue();
            break;
        default:
            break;
    }
}
