/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/*          EXIFPROBE - TIFF/JPEG/EXIF image file probe               */
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */
/* Copyright (C) 2002 by Duane H. Hesser. All rights reserved.        */
/*                                                                    */
/* See the file LICENSE.EXIFPROBE for terms of use.                   */
/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */

#ifndef lint
static char *ModuleId = "@(#) $Id: process.c,v 1.26 2003/05/08 17:23:48 alex Exp $";
#endif

/* Process segment types such as TIFF IFD, JPEG APP segments, and     */
/* JPEG basic segments.                                               */

#include <sys/types.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "defs.h"
#include "extern.h"
#include "misc.h"
#include "tags.h"

/* Read and decode a TIFF IFD, attempting to describe the location    */
/* and structure of the data as it is read. Entry values which live   */
/* at an offset from the entry are displayed following the last entry */
/* in the IFD (actually after the "next ifd offset") if               */
/* VALUE_AT_OFFSET is set in Print_options, so that most IFDS will    */
/* have a VALUES section following the entry lines, in just the       */
/* manner that the file is laid out by spec. Tag names for offset     */
/* values are repeated when the value is printed. If VALUE_AT_OFFSET  */
/* is not set, the value is printed immediately, unless the tag       */
/* represents a subsection, in which case the offset of the subsect   */
/* in printed as the tag value, and the subsection handled in a       */
/* second pass, after all entries are done.                           */

/* Returns a file offset which represents the "highest" position of   */
/* the file pointer read by this routine and its subroutines, which   */
/* may be used by the caller to read the remainder of the file. This  */
/* number (less one) is printed as the offset of the end of the IFD.  */

/* At the same time, the function notices non-zero values of "next    */
/* ifd offset" which may be contained in this ifd, and passes it as   */
/* "max_offset" to routines which read subifds, on the assumption     */
/* that this ifd should not write into the next ifd. That value is    */
/* used to mark segments which appear to be out of order in the file. */
/* If the routine detects by this means that *this* ifd is outside    */
/* its natural bounds, and cannot be contained within its parent, it  */
/* returns 0, rather than the maximum value offset, so that the       */
/* parent routine will not improperly alter its maximum offset.       */

/* The arguments are the byteorder to be used when reading integer    */
/* data (usually taken from the TIFF header), the offset of the start */
/* of the directory, relative to the TIFF header, the offset of the   */
/* TIFF header from the beginning of the file, the sequence number of */
/* the IFD, and formatting values (indent and address width) to be    */
/* used when printing.                                                */

/* The routine also attempts to record the location, size, and        */
/* compression method used for thumbnail and main image data for TIFF */
/* format files, and the location and size of thumbnail data for JPEG */
/* interchange formats which store auxilliary information in TIFF     */
/* format IFDs (and thus call this routine). This information is      */
/* printed at the end, which is normally the end of the "auxilliary"  */
/* data (for single main image files), and will be followed by the    */
/* thumbnail/main-image data.                                         */

unsigned long
process_tiff_ifd(FILE *inptr,unsigned short byteorder,unsigned long ifd_offset,
                    unsigned long fileoffset_base,unsigned long max_offset,
                    int ifdtype,int ifdnum,int subifdnum,int indent,
                    int addrwidth)
{
    struct ifd_entry *entry_ptr;
    struct tiff_image_summary tif_img[MAXIFDS];
    unsigned long cur_ifd_offset,next_ifd_offset,current_offset;
    unsigned long start_entry_offset,entry_offset,value_offset;
    unsigned long max_value_offset = 0UL;
    unsigned long max_ifd_offset = 0UL;
    unsigned long thn_offset = 0UL;          /* For JPEG format       */
    unsigned long thn_length = 0UL;          /* For JPEG format       */
    unsigned short marker;
    int value_is_offset = 0;
    int use_second_pass = 0;
    int nifds = -1;
    int entry_num,num_entries,i;
    char *ifdname;

    if(inptr == (FILE *)0)
    {
        fprintf(stderr,"%s: no open file pointer to read TIFF IFD\n",
                Progname);
        return(0L);
    }

    next_ifd_offset = ifd_offset + fileoffset_base;
    while(next_ifd_offset)
    {
        cur_ifd_offset = next_ifd_offset;
        if(max_offset && (cur_ifd_offset > max_offset))
            print_tag_address(SECTION,cur_ifd_offset,indent,"^",addrwidth);
        else
            print_tag_address(SECTION,cur_ifd_offset,indent,"@",addrwidth);

        switch(ifdtype)
        {
            case TIFF_IFD:
                if(PRINT_SECTION)
                    printf("<IFD%d>",ifdnum);
                ifdname = "IFD";
                break;
            case TIFF_SUBIFD:
                if(PRINT_SECTION)
                    printf("<SubIFD%d%d>",ifdnum,subifdnum);
                ifdname = "SubIFD";
                break;
            case INTEROP_IFD:
                if(PRINT_SECTION)
                    printf("<Interoperability SubIFD>");
                ifdname = "Interoperability SubIFD";
                break;
            case GPS_IFD:
                if(PRINT_SECTION)
                    printf("<GPS SubIFD>");
                ifdname = "GPS SubIFD";
                break;
            case EXIF_IFD:  /* This shouldn't happen                  */
                if(PRINT_SECTION)
                    printf("<EXIF IFD>");   
                ifdname = "EXIF IFD";
                break;
            default:
                if(PRINT_SECTION)
                    printf("<UNKNOWN IFD TYPE %d>",ifdtype);    /* SECTION */
                ifdname = "UNKNOWN IFD TYPE";
                break;
        }

        ++nifds;
        current_offset = ftell(inptr);
        num_entries = read_ushort(inptr,byteorder,cur_ifd_offset);
        if(ferror(inptr) || feof(inptr))
        {
            printred("#========= WARNING: FAILED to read next IFD at offset ");
            printf("%lu\n",cur_ifd_offset);
            goto blewit;
        }
        current_offset = entry_offset = start_entry_offset = ftell(inptr);
        if(PRINT_SECTION)
        {
            printf(" %d entries ",num_entries);
            printf("starting at file offset %#lx=%lu\n",
                    start_entry_offset,start_entry_offset);
        }

        tif_img[nifds].offset = tif_img[nifds].size = 0L;
        tif_img[nifds + 1].offset = tif_img[nifds + 1].size = 0L;
        tif_img[nifds].compression = tif_img[nifds + 1].compression = 0;
        tif_img[nifds].type = tif_img[nifds + 1].type = UNKNOWN_TYPE;

        use_second_pass = value_is_offset = 0;

        for(entry_num = 0; entry_num < num_entries; ++entry_num)
        {
            entry_ptr = read_entry(inptr,byteorder,entry_offset);
            if((entry_ptr->value_type == 0) || (entry_ptr->value_type > DOUBLE) ||
                ferror(inptr) || feof(inptr))
            {
                print_tag_address(ENTRY,entry_offset,indent,"@",addrwidth);
                print_taginfo(entry_ptr,SUBINDENT,ifdtype,ifdnum,subifdnum);
                printred(" INVALID ENTRY\n");
                goto blewit;
            }
            current_offset = ftell(inptr);

            switch(entry_ptr->tag)
            {
                case TIFFTAG_SubFileType:
                    if(entry_ptr->value == 1)
                        tif_img[nifds].type = PRIMARY_TYPE;
                    else if(entry_ptr->value == 2)
                        tif_img[nifds].type = THUMBNAIL_TYPE;
                    else if(entry_ptr->value == 3)
                        tif_img[nifds].type = PAGE_TYPE;
                    break;
                case TIFFTAG_NewSubFileType:
                    if(entry_ptr->value == 0)
                        tif_img[nifds].type = PRIMARY_TYPE;
                    else if(entry_ptr->value == 1)
                        tif_img[nifds].type = THUMBNAIL_TYPE;
                    else if(entry_ptr->value == 2)
                        tif_img[nifds].type = PAGE_TYPE;
                    else if(entry_ptr->value == 3)
                        tif_img[nifds].type = MASK_TYPE;
                    break;
                case TIFFTAG_Compression:
                    tif_img[nifds].compression = entry_ptr->value;
                    break;
                case TIFFTAG_JPEGInterchangeFormat:
                    /* actually an offset...                          */
                    thn_offset = entry_ptr->value + fileoffset_base;
                    break;
                case TIFFTAG_JPEGInterchangeFormatLength:
                    thn_length = entry_ptr->value;
                    break;
                case TIFFTAG_StripOffsets:
                case TIFFTAG_TileOffsets:
                    tif_img[nifds].offset = entry_ptr->value;
                    break;
                case TIFFTAG_StripByteCounts:
                case TIFFTAG_TileByteCounts:
                    tif_img[nifds].size = entry_ptr->value;
                    break;
                case TIFFTAG_SubIFDtag:
                case EXIFTAG_ExifIFDPointer:
                case EXIFTAG_GPSInfoIFDPointer:
                case EXIFTAG_Interoperability: 
                    use_second_pass++;
                    break;
                default:
                    break;
            }
            print_tag_address(ENTRY,entry_offset,indent,"@",addrwidth);
            value_offset = print_entry(inptr,byteorder,addrwidth,ifdtype,ifdnum,
                                        subifdnum,entry_ptr,fileoffset_base,SUBINDENT);
            if(value_offset == 0UL)
                value_offset = current_offset;
            if(value_offset > max_value_offset)
                max_value_offset = value_offset;
            if((PRINT_VALUE_AT_OFFSET) && (is_offset(entry_ptr)))
                ++use_second_pass;
            entry_offset = current_offset;
        }

        next_ifd_offset = read_ulong(inptr,byteorder,current_offset);
        if(next_ifd_offset > 0UL)
        {
            if((PRINT_ADDRESS) && (PRINT_VALUE) && (PRINT_VALUE_AT_OFFSET))
            {
                print_tag_address(ENTRY,current_offset,indent,"@",addrwidth);
                extraindent(SUBINDENT);
                printf("**** next IFD offset %lu  ",next_ifd_offset);
                next_ifd_offset += fileoffset_base;
                printf("(+ %lu = %#lx/%lu)",fileoffset_base,
                            next_ifd_offset,next_ifd_offset);
                if((next_ifd_offset < ftell(inptr)) ||
                    ((ifdtype != TIFF_IFD) && (ifdtype != TIFF_SUBIFD)))
                {
                    printred(" INVALID NEXT IFD OFFSET\n");
                    next_ifd_offset = 0UL;
                }
                else putchar('\n');
            }
            else
            {
                if((next_ifd_offset < ftell(inptr)) || 
                    ((ifdtype != TIFF_IFD) && (ifdtype != TIFF_SUBIFD)))
                {
                    printred("#========= WARNING: INVALID NEXT IFD OFFSET in ");
                    if(ifdtype == TIFF_IFD)
                        printf("IFD %d =========\n",ifdnum);
                    else if(ifdtype == TIFF_SUBIFD)
                        printf("IFD %d%d =========\n",ifdnum,subifdnum);
                    else
                        printf("%s =========\n",ifdname);
                    next_ifd_offset = 0L;
                }
                else
                    next_ifd_offset += fileoffset_base;
            }
            /* offsets found in the next pass should be within the    */
            /* bounds of the ifd. The following helps to detect       */
            /* chunks that are written "out of place". "max_offset"   */
            /* will be 0 unless this is a subifd, in which case it    */
            /* will be the limit for the parent. If "next_ifd_offset" */
            /* is non-zero, this ifd should be done by then. This     */
            /* value is set to the lesser non-zero value.             */
            max_ifd_offset = next_ifd_offset;
            if(max_offset && (max_ifd_offset > max_offset))
                max_ifd_offset = max_offset;
        }
        else
        {
            max_ifd_offset = max_offset;    
            if((PRINT_ADDRESS) && (PRINT_VALUE) && (PRINT_VALUE_AT_OFFSET))
            {
                print_tag_address(ENTRY,current_offset,indent,"@",addrwidth);
                extraindent(SUBINDENT);
                printf("**** next IFD offset 0\n");
            }
        }

        /* If we made it through the first pass, we should be able to */
        /* get some info from the second pass, so don't let a failure */
        /* here stop us. It will fail later.                          */
        if(ferror(inptr) || feof(inptr))
        {
            printf("#========= WARNING: READ NEXT IFD OFFSET FAILED =========\n");
            why(stdout);
            clearerr(inptr);    /* keep going...                      */
        }
        else
            current_offset = ftell(inptr);

        value_offset = current_offset;
        if(use_second_pass)
        {
            if((PRINT_VALUE) && (PRINT_ADDRESS) && (PRINT_VALUE_AT_OFFSET))
            {
                print_tag_address(ENTRY,value_offset,indent,"@",addrwidth);
                extraindent(SUBINDENT);
                printf("============= VALUES, ");

                if(ifdtype == TIFF_IFD)
                    printf("IFD%d",ifdnum);
                else if(ifdtype == TIFF_SUBIFD)
                    printf("SubIFD%d%d",ifdnum,subifdnum);
                else if(ifdtype == INTEROP_IFD)
                    printf("Interoperability SubIFD");
                else if(ifdtype == EXIF_IFD)    /* shouldn't happen   */
                    printf("EXIF IFD");
                else
                    printf("UNKNOWN IFD TYPE %d",ifdtype);
                printf(" ============\n");
            }

            /* Second pass, to evaluate entries which are stored      */
            /* indirectly (the value requires more than 4 bytes).     */
            /* Also grab compression, image offset and size, and      */
            /* thumbnail information on this pass, and process Exif,  */
            /* Interop, and SubIfds.                                  */

            /* This time we have to explicitly seek to each entry,    */
            /* since the value processing may send the file pointer   */
            /* off to exotic places.                                  */
            entry_offset = start_entry_offset;
            for(entry_num = 0; entry_num < num_entries; ++entry_num)
            {
                entry_ptr = read_entry(inptr,byteorder,entry_offset);
                if((entry_ptr->value_type == 0) || (entry_ptr->value_type > DOUBLE) ||
                    ferror(inptr) || feof(inptr))
                {
                    print_tag_address(ENTRY,entry_offset,indent,"@",addrwidth);
                    printf(" INVALID (%lu)\n",entry_ptr->value);
                    goto blewit;
                }
                entry_offset = current_offset = ftell(inptr);
                value_is_offset = is_offset(entry_ptr);
                switch(entry_ptr->tag)
                {
                    case TIFFTAG_SubIFDtag:
                        value_offset = process_subifd(inptr,byteorder,
                                            entry_ptr,fileoffset_base,
                                            max_ifd_offset,nifds,
                                            TIFF_SUBIFD,indent,addrwidth);
                        break;
                    case EXIFTAG_ExifIFDPointer:
                        value_offset = process_exif_ifd(inptr,byteorder,
                                            entry_ptr->value,fileoffset_base,
                                            max_ifd_offset,nifds,
                                            indent+SUBINDENT,addrwidth);
                        break;
                    case EXIFTAG_GPSInfoIFDPointer:
                        value_offset = process_gps_ifd(inptr,byteorder,
                                            entry_ptr->value,fileoffset_base,
                                            max_ifd_offset,nifds,
                                            indent+SUBINDENT,addrwidth);
                        break;
                    case EXIFTAG_Interoperability: 
                        /* This doesn't belong in a TIFF IFD, but     */
                        /* "gexif" will inexplicably move it here, so */
                        /* be prepared.                               */
                        STARTCOLOR(INTEROP_COLOR);
                        value_offset = process_subifd(inptr,byteorder,
                                        entry_ptr,fileoffset_base,
                                        max_ifd_offset,-1,INTEROP_IFD,
                                        indent,addrwidth);
                        STOPCOLOR;
                        break;
                    case TIFFTAG_StripOffsets:
                    case TIFFTAG_TileOffsets:
                        if(value_is_offset)
                        {
                            if(entry_ptr->value_type == LONG)
                                tif_img[nifds].offset =
                                    read_ulong(inptr,byteorder,
                                        entry_ptr->value + fileoffset_base);
                            else
                                tif_img[nifds].offset =
                                    read_ushort(inptr,byteorder,
                                        entry_ptr->value + fileoffset_base);
                        }
                        if((PRINT_VALUE) && ((PRINT_VALUE_AT_OFFSET) && value_is_offset))
                        {
                            print_tag_address(ENTRY,fileoffset_base + entry_ptr->value,
                                                indent,"@",addrwidth);
                            print_tagid(entry_ptr,SUBINDENT,EXIF_IFD);
                            value_offset = 
                                print_offset_value(inptr,byteorder,
                                            entry_ptr,fileoffset_base,
                                            ifdtype,indent,addrwidth,1);
                        }
                        break;
                    case TIFFTAG_StripByteCounts:
                    case TIFFTAG_TileByteCounts:
                        if(value_is_offset)
                        {
                            tif_img[nifds].size =
                                sum_strip_bytecounts(inptr,byteorder,
                                    entry_ptr->value + fileoffset_base,
                                    entry_ptr->count,entry_ptr->value_type);
                        }
                        if((PRINT_VALUE) && ((PRINT_VALUE_AT_OFFSET) && value_is_offset))
                        {
                            print_tag_address(ENTRY,fileoffset_base + entry_ptr->value,
                                                indent,"@",addrwidth);
                            print_tagid(entry_ptr,SUBINDENT,EXIF_IFD);
                            value_offset =
                                print_offset_value(inptr,byteorder,
                                            entry_ptr,fileoffset_base,
                                            ifdtype,indent,addrwidth,1);
                        }
                        break;
                    default:
                        if((PRINT_VALUE) && ((PRINT_VALUE_AT_OFFSET) && value_is_offset))
                        {
                            print_tag_address(ENTRY,fileoffset_base + entry_ptr->value,
                                                    indent,"@",addrwidth);
                            print_tagid(entry_ptr,SUBINDENT,ifdtype);
                            value_offset =
                                print_offset_value(inptr,byteorder,
                                        entry_ptr,fileoffset_base,
                                        ifdtype,indent,addrwidth,1);
                        }
                        break;
                }
                if(value_offset == 0L) /* bad entry; try the next one */
                    clearerr(inptr);
                else if(value_offset > max_value_offset)
                    max_value_offset = value_offset;
            }
        }
        else if(current_offset > max_value_offset)
            max_value_offset = current_offset;

        max_ifd_offset = max_value_offset;
        if(max_offset && (cur_ifd_offset > max_offset))
        {
            print_tag_address(SECTION,max_ifd_offset - 1,indent,"*",addrwidth);
            max_ifd_offset = 0L;
        }
        else
            print_tag_address(SECTION,max_ifd_offset - 1,indent,"-",addrwidth);

        /* Mark the end of the IFD                                    */
        switch(ifdtype)
        {
            case TIFF_IFD:
                if(PRINT_SECTION)
                    printf("</IFD%d>\n",ifdnum);
                ifdnum++;
                break;
            case TIFF_SUBIFD:
                if(PRINT_SECTION)
                    printf("</SubIFD%d%d>\n",ifdnum,subifdnum);
                subifdnum++;
                break;
            case INTEROP_IFD:
                if(PRINT_SECTION)
                    printf("</Interoperability SubIFD>\n");
                break;
            case GPS_IFD:
                if(PRINT_SECTION)
                    printf("</GPS SubIFD>\n");
                break;
            case EXIF_IFD:
                if(PRINT_SECTION)
                    printf("</EXIF IFD>\n");
                break;
            default:
                if(PRINT_SECTION)
                    printf("</UNKNOWN IFD TYPE %d>\n",ifdtype);
                break;
        }

        if(thn_offset)
        {
            /* This will be set only if a JpegInterchangeFormat tag   */
            /* has been seen, so this must be a JPEG thumbnail.       */
            print_tag_address(SECTION,thn_offset,indent,"@",addrwidth);
            if(PRINT_SECTION)
                printf("#### Start of JPEG thumbnail data, length %lu ####\n",thn_length);
            marker = read_imagemarker(inptr,thn_offset);
            if(marker)
                max_value_offset = process_jpeg_segments(inptr,thn_offset,marker,
                                            thn_length,indent,
                                            addrwidth);

            if(PRINT_SECTION)
            {
                print_tag_address(SECTION,thn_offset + thn_length - 1,indent,"-",addrwidth);
                printf("#### End of JPEG thumbnail data ####");
                if(max_value_offset == 0L)
                    printf(" JPEG_EOI NOT FOUND");
                (void)putchar('\n');
            }
        }
        /* If there is a next ifd, "max_offset" must be reset.        */
        max_offset = 0L;
    }

    if(PRINT_SECTION)
    {
        for(i = 0; i <= nifds; i += 2)
        {
            int j;
            unsigned long other_size;

            if(nifds > 0)
            {
                if(i % 2)
                    j = i - 1;
                else
                    j = i + 1;
                if(tif_img[i].offset > tif_img[j].offset)
                {
                    STARTCOLOR(IM_COLOR);
                    other_size = tif_img[i].size;
                    print_tif_summary(&tif_img[j],other_size,fileoffset_base,
                                        ifdname,j,indent,"@",addrwidth);
                    other_size = tif_img[j].size;
                    print_tif_summary(&tif_img[i],other_size,fileoffset_base,
                                        ifdname,i,indent,"@",addrwidth);
                    STOPCOLOR;
                }
                else if(tif_img[i].size)
                {
                    STARTCOLOR(IM_COLOR);
                    other_size = tif_img[j].size;
                    print_tif_summary(&tif_img[i],other_size,fileoffset_base,
                                        ifdname,i,indent,"@",addrwidth);
                    other_size = tif_img[i].size;
                    print_tif_summary(&tif_img[j],other_size,fileoffset_base,
                                        ifdname,j,indent,"@",addrwidth);
                    STOPCOLOR;
                }
            }
            else if(tif_img[i].size)
            {
                STARTCOLOR(IM_COLOR);
                print_tif_summary(&tif_img[i],0L,fileoffset_base,ifdname,i,
                                    indent,"@",addrwidth);
                STOPCOLOR;
            }
        }
    }
    return(max_ifd_offset);
blewit:
    clearerr(inptr);
    if(PRINT_SECTION)
    {
        /* Indicate the end of the IFD                                    */
        current_offset = ftell(inptr);
        print_tag_address(SECTION,current_offset - 1,indent,"#",addrwidth);
        switch(ifdtype)
        {
            case TIFF_IFD:
                printf("</IFD%d>\n",ifdnum);
                break;
            case TIFF_SUBIFD:
                printf("</SubIFD%d%d>\n",ifdnum,subifdnum);
                break;
            case INTEROP_IFD:
                printf("</Interoperability SubIFD>\n");
                break;
            case GPS_IFD:
                printf("</GPS SubIFD>\n");
                break;
            case EXIF_IFD:
                if(PRINT_SECTION)
                    printf("</EXIF IFD>\n");
                break;
            default:
                printf("</UNKNOWN IFD TYPE %d>\n",ifdtype);
                break;
        }
    }
    return(0L);
}

/* Read, decode, and print an EXIF IFD at 'exif_offset', noticing     */
/* MakerNotes and "Interoperability" IFDs as they go by. The offset   */
/* of each entry and offset value is recorded and printed, and the    */
/* offset of the last byte read is reported as the end of the IFD. If */
/* all goes well, the offset at which the caller should next read     */
/* (just past the end of the offset values) is returned. If an        */
/* unrecoverable error occurs, 0UL is returned.                       */

/* This looks remoarkably like process_tiff_ifd(), except that it     */
/* expects EXIF tags, and doesn't expect further IFDs to be chained   */
/* from it. If it does find a chained IFD, it processes it as a       */
/* normal IFD (not another EXIF IFD). It should probably just warn    */
/* and ignore it.                                                     */

unsigned long
process_exif_ifd(FILE *inptr,unsigned short byteorder,
                    unsigned long exif_offset,
                    unsigned long fileoffset_base,
                    unsigned long max_offset,
                    int ifdnum,int indent,int addrwidth)
{
    struct ifd_entry *entry_ptr;
    unsigned long max_value_offset = 0L;
    unsigned long next_ifd_offset,current_offset,max_exif_offset;
    unsigned long start_entry_offset,entry_offset,value_offset;
    int entry_num,num_entries,value_is_offset,use_second_pass;

    if(inptr == (FILE *)0)
    {
        fprintf(stderr,"%s: no open file pointer to read EXIF IFD\n",
                Progname);
        return(0L);
    }

    STARTCOLOR(EXIF_COLOR);
    /* If the exif segment appears to be beyond the end of the parent */
    /* ifd, mark the start address to call attention.                 */
    if(max_offset && (exif_offset > max_offset))
        print_tag_address(SECTION,exif_offset + fileoffset_base,indent,"^",addrwidth);
    else
        print_tag_address(SECTION,exif_offset + fileoffset_base,indent,"@",addrwidth);
    if(PRINT_SECTION)
        printf("<EXIF IFD> (in IFD %d)",ifdnum);

    num_entries = read_ushort(inptr,byteorder,exif_offset + fileoffset_base);
    if(ferror(inptr) || feof(inptr))
        goto blewit;
    current_offset = entry_offset = start_entry_offset = ftell(inptr);
    if(PRINT_SECTION)
    {
        printf(" %d entries ",num_entries);
        printf("starting at file offset %#lx=%lu\n",
                current_offset,current_offset);
    }
    use_second_pass = value_is_offset = 0;

    /* The direct entries                                             */
    for(entry_num = 0; entry_num < num_entries; ++entry_num)
    {
        entry_ptr = read_entry(inptr,byteorder,entry_offset);
        if((entry_ptr->value_type == 0) || (entry_ptr->value_type > DOUBLE) ||
            ferror(inptr) || feof(inptr))
        {
            print_tag_address(ENTRY,entry_offset,indent,"@",addrwidth);
            printf(" INVALID EXIF ENTRY (%lu)\n",entry_ptr->value);
            goto blewit;
        }
        current_offset = ftell(inptr);
        switch(entry_ptr->tag)
        {
            case EXIFTAG_Interoperability: 
            case EXIFTAG_MakerNote: 
                use_second_pass++;
                break;
            default:
                break;
        }
     
        print_tag_address(ENTRY,entry_offset,indent,"@",addrwidth);
        value_offset = print_entry(inptr,byteorder,addrwidth,EXIF_IFD,ifdnum,
                                        -1,entry_ptr,fileoffset_base,SUBINDENT);
        if(value_offset == 0UL)
            value_offset = current_offset;
        if(value_offset > max_value_offset)
            max_value_offset = value_offset;
        if((PRINT_VALUE_AT_OFFSET) && (is_offset(entry_ptr)))
            ++use_second_pass;
        entry_offset = current_offset;
    }

    next_ifd_offset = read_ulong(inptr,byteorder,current_offset);

    /* offsets found in the next pass should be within the bounds of  */
    /* the ifd. The following helps to detect chunks that are written */
    /* "out of place".                                                */
    if(next_ifd_offset > 0L)
    {
        if((PRINT_ADDRESS) && (PRINT_VALUE) && (PRINT_VALUE_AT_OFFSET))
        {
            print_tag_address(ENTRY,current_offset,indent,"@",addrwidth);
            extraindent(SUBINDENT);
            printf("**** next IFD offset %lu  ",next_ifd_offset);
            next_ifd_offset += fileoffset_base;
            if(next_ifd_offset < ftell(inptr))
            {
                printred("BAD NEXT IFD OFFSET\n");
                next_ifd_offset = 0L;
            }
            else
                printf("(+ %lu = %#lx/%lu)\n",fileoffset_base,
                                next_ifd_offset,next_ifd_offset);
        }
        else
            next_ifd_offset += fileoffset_base;
        /* This should never happen                                   */
        max_exif_offset = next_ifd_offset;
        if(max_offset && (max_exif_offset > max_offset))
            max_exif_offset = max_offset;
    }
    else
    {
        max_exif_offset = max_offset;   
        if((PRINT_ADDRESS) && (PRINT_VALUE) && (PRINT_VALUE_AT_OFFSET))
        {
            print_tag_address(ENTRY,current_offset,indent,"@",addrwidth);
            extraindent(SUBINDENT);
            printf("**** next IFD offset 0\n");
        }
    }

    if(ferror(inptr) || feof(inptr))
    {
        printf(" READ NEXT IFD OFFSET FAILED ");
        why(stdout);
        clearerr(inptr);    /* keep going...                          */
    }
    else
        current_offset = ftell(inptr);

    value_offset = current_offset;

    /* Second pass, to evaluate entries which are stored indirectly.  */
    /* This occurs when the value requires more than 4 bytes, AND for */
    /* certain LONG values which are intended to be used as offsets   */
    /* (and should have their own type defined in the spec),          */
    if(use_second_pass)
    {
        if((PRINT_VALUE) && (PRINT_ADDRESS) && (PRINT_VALUE_AT_OFFSET))
        {
            print_tag_address(ENTRY,value_offset,indent,"@",addrwidth);
            extraindent(SUBINDENT);
            printf("============= VALUES, EXIF IFD ============\n");
        }

        entry_offset = start_entry_offset;
        for(entry_num = 0; entry_num < num_entries; ++entry_num)
        {
            entry_ptr = read_entry(inptr,byteorder,entry_offset);
            if((entry_ptr->value_type == 0) || (entry_ptr->value_type > DOUBLE) ||
                ferror(inptr) || feof(inptr))
            {
                print_tag_address(ENTRY,entry_offset,indent,"@",addrwidth);
                printf(" INVALID (%lu)\n",entry_ptr->value);
                goto blewit;
            }
            current_offset = entry_offset = ftell(inptr);
            switch(entry_ptr->tag)
            {
                case EXIFTAG_Interoperability: 
                    STARTCOLOR(INTEROP_COLOR);
                    value_offset = process_subifd(inptr,byteorder,
                                    entry_ptr,fileoffset_base,
                                    max_exif_offset,-1,INTEROP_IFD,
                                    indent,addrwidth);
                    STOPCOLOR;
                    break;
                case EXIFTAG_MakerNote: 
                    if((PRINT_SECTION) || (PRINT_VALUE))
                        value_offset = process_makernote(inptr,byteorder,
                                    entry_ptr,fileoffset_base,
                                    max_exif_offset,indent,addrwidth);
                    break;
                default:
                    if((PRINT_VALUE) && ((PRINT_VALUE_AT_OFFSET) && is_offset(entry_ptr)))
                    {
                        print_tag_address(ENTRY,fileoffset_base + entry_ptr->value,
                                            indent,"@",addrwidth);
                        print_tagid(entry_ptr,SUBINDENT,EXIF_IFD);
                        value_offset =
                                    print_offset_value(inptr,byteorder,
                                        entry_ptr,fileoffset_base,
                                        EXIF_IFD,indent,addrwidth,1);
                    }
                    break;
            }
            if(value_offset > max_value_offset)
                max_value_offset = value_offset;
        }
    }
    else if(current_offset > max_value_offset)
        max_value_offset = current_offset;

    if(max_offset && (exif_offset > max_offset))
    {
        /* The Exif IFD is (still) somehow outside the parent IFD     */
        print_tag_address(SECTION,max_value_offset - 1,indent,"*",addrwidth);
        max_value_offset = 0L;  /* let the caller know.               */
    }
    else
        print_tag_address(SECTION,max_value_offset - 1,indent,"-",addrwidth);
    if(PRINT_SECTION)
        printf("</EXIF IFD>\n");
    STOPCOLOR;
    /* This shouldn't happen, but considering the disdain for         */
    /* specifications displayed by camera makers so far, well, let's  */
    /* just see...    then there's also corruption caused by editors  */
    if(next_ifd_offset)
    {
        printred("#========= WARNING! CHAINING NEXT IFD FROM EXIF IFD =========\n");
        value_offset = process_tiff_ifd(inptr,byteorder,next_ifd_offset,
                                fileoffset_base,max_offset, EXIF_IFD,
                                ++ifdnum,-1,indent,addrwidth);
        if(value_offset > max_value_offset)
            max_value_offset = value_offset;
    }
    return(max_value_offset);
blewit:
    /* ###%%% report current_offset, why()                            */
    clearerr(inptr);
    current_offset = ftell(inptr);
    print_tag_address(SECTION,current_offset - 1,indent,"-",addrwidth);
    if(PRINT_SECTION)
        printf("</EXIF IFD>\n");
    STOPCOLOR;
    return(0L);
}

/* Read an ifd entry describing one or more subifd offsets, and       */
/* process the IFDs found at those offsets as TIFF image file         */
/* directories. This is used for TIFF6/TIFFEP SubIFDtag and EXIF      */
/* Interoperability tag. "ifdtype" will normally be "SubIFD" or       */
/* "Interoperability SubIFD". 'ifdnum" will be -1 for                 */
/* Interoperability subifd, >= 0 for tagged SubIFDs.                  */

/* The return will be the position of the file pointer when done      */
/* reading the subifd, which should be the offset just past the end   */
/* of the subifd. The return is 0L if something went wrong.           */

unsigned long
process_subifd(FILE *inptr,unsigned short byteorder,
                    struct ifd_entry *subfile_entry,
                    unsigned long fileoffset_base,
                    unsigned long max_offset,
                    int ifdnum,int ifdtype,int indent,int addrwidth)
{
    unsigned long entry_offset;
    unsigned long subifd_offset;
    unsigned long max_value_offset = 0L;
    int subifdnum = 0;
    int nsubifds;

    if(inptr && subfile_entry)
    {
        entry_offset = subfile_entry->value + fileoffset_base;
        nsubifds = subfile_entry->count;

        if(ifdtype != INTEROP_IFD)
            ifdtype = TIFF_SUBIFD;
        else
            subifdnum = -1;
        if(nsubifds > 1)
        {
            if((PRINT_VALUE_AT_OFFSET))
            {
                if(max_offset && (entry_offset > max_offset))
                    print_tag_address(SECTION,entry_offset,indent,"^",addrwidth);
                else
                    print_tag_address(SECTION,entry_offset,indent,"@",addrwidth);
                if(PRINT_SECTION)
                {
                    extraindent(SUBINDENT);
                    printf("%s - %lu offsets: [max parent offset = %lu]",
                            tagname(subfile_entry->tag),subfile_entry->count,
                            max_offset);
                }
                for(subifdnum = 0L; subifdnum < nsubifds; ++subifdnum)
                {
                    subifd_offset = read_ulong(inptr,byteorder,entry_offset)
                                                        + fileoffset_base;
                    entry_offset = ftell(inptr); /* where to get next one */
                    printf(" %lu",subifd_offset);
                    if(ferror(inptr) || feof(inptr))
                        goto blewit;
                }
                if(PRINT_SECTION)
                    (void)putchar('\n');
            }

            entry_offset = subfile_entry->value + fileoffset_base;
            for(subifdnum = 0L; subifdnum < nsubifds; ++subifdnum)
            {
                subifd_offset = read_ulong(inptr,byteorder,entry_offset) +
                                                    fileoffset_base;
                entry_offset = ftell(inptr); /* where to get next one */
                if(ferror(inptr) || feof(inptr))
                    goto blewit;
                max_value_offset = process_tiff_ifd(inptr,byteorder,
                                        subifd_offset,fileoffset_base,
                                        max_offset,ifdtype,ifdnum,
                                        subifdnum,indent+SUBINDENT,
                                        addrwidth);
                /* ###%%% don't we check for errors here???           */
            }
        }
        else
        {
            max_value_offset = process_tiff_ifd(inptr,byteorder,
                                        subfile_entry->value,
                                        fileoffset_base,max_offset,
                                        ifdtype,ifdnum,subifdnum,
                                        indent+SUBINDENT, addrwidth);
        }
    }
    else
    {
        fprintf(stderr,"%s: no open file pointer to read EXIF\n",
                Progname);
        max_value_offset = 0L;
    }
    return(max_value_offset);
blewit:
    /* we're going to keep trying, because the caller's next offset   */
    /* may be ok.                                                     */
    clearerr(inptr);
    return(0L);
}

/* Just a bare bones GPS routine; no interpretation is done.          */

unsigned long
process_gps_ifd(FILE *inptr,unsigned short byteorder,
                    unsigned long gps_offset,
                    unsigned long fileoffset_base,
                    unsigned long max_offset,
                    int ifdnum,int indent,int addrwidth)
{
    struct ifd_entry *entry_ptr;
    unsigned long max_value_offset = 0L;
    unsigned long next_ifd_offset,current_offset,max_gps_offset;
    unsigned long start_entry_offset,entry_offset,value_offset;
    int entry_num,num_entries,value_is_offset,use_second_pass;

    if(inptr == (FILE *)0)
    {
        fprintf(stderr,"%s: no open file pointer to read GPS IFD\n",
                Progname);
        return(0L);
    }

    STARTCOLOR(GPS_COLOR);
    /* If the gps segment appears to be beyond the start of the       */
    /* parent ifd, mark the start address funny.                      */
    if(max_offset && (gps_offset > max_offset))
        print_tag_address(SECTION,gps_offset + fileoffset_base,indent,"^",addrwidth);
    else
        print_tag_address(SECTION,gps_offset + fileoffset_base,indent,"@",addrwidth);
    if(PRINT_SECTION)
        printf("<GPS IFD> (in IFD %d)",ifdnum);

    num_entries = read_ushort(inptr,byteorder,gps_offset + fileoffset_base);
    if(ferror(inptr) || feof(inptr))
        goto blewit;
    current_offset = entry_offset = start_entry_offset = ftell(inptr);
    if(PRINT_SECTION)
    {
        printf(" %d entries ",num_entries);
        printf("starting at file offset %#lx=%lu\n",
                current_offset,current_offset);
    }
    use_second_pass = value_is_offset = 0;

    /* The direct entries                                             */
    for(entry_num = 0; entry_num < num_entries; ++entry_num)
    {
        print_tag_address(ENTRY,entry_offset,indent,"@",addrwidth);
        entry_ptr = read_entry(inptr,byteorder,entry_offset);
        if((entry_ptr->value_type == 0) || (entry_ptr->value_type > DOUBLE) ||
            ferror(inptr) || feof(inptr))
        {
            print_tag_address(ENTRY,entry_offset,indent,"@",addrwidth);
            printf(" INVALID (%lu)\n",entry_ptr->value);
            goto blewit;
        }
        entry_offset = current_offset = ftell(inptr);
        value_offset = print_entry(inptr,byteorder,addrwidth,GPS_IFD,ifdnum,
                                    -1,entry_ptr,fileoffset_base,SUBINDENT);
        if(value_offset == 0UL)
            value_offset = entry_offset;
        if(value_offset > max_value_offset)
            max_value_offset = value_offset;
        if((PRINT_VALUE_AT_OFFSET) && (is_offset(entry_ptr)))
            ++use_second_pass;
    }

    next_ifd_offset = read_ulong(inptr,byteorder,current_offset);

    /* offsets found in the next pass should be within the bounds of  */
    /* the ifd. The following helps to detect chunks that are written */
    /* "out of place".                                                */
    if(next_ifd_offset > 0L)
    {
        if((PRINT_ADDRESS) && (PRINT_VALUE) && (PRINT_VALUE_AT_OFFSET))
        {
            print_tag_address(ENTRY,current_offset,indent,"@",addrwidth);
            extraindent(SUBINDENT);
            printf("**** next IFD offset %lu  ",next_ifd_offset);
            next_ifd_offset += fileoffset_base;
            if(next_ifd_offset < ftell(inptr))
            {
                printred("BAD NEXT IFD OFFSET\n");
                next_ifd_offset = 0L;
            }
            else
                printf("(+ %lu = %#lx/%lu)\n",fileoffset_base,
                            next_ifd_offset,next_ifd_offset);
        }
        else
            next_ifd_offset += fileoffset_base;
        /* This should never happen                                   */
        max_gps_offset = next_ifd_offset;
        if(max_offset && (max_gps_offset > max_offset))
            max_gps_offset = max_offset;
    }
    else
    {
        max_gps_offset = max_offset;    
        if((PRINT_ADDRESS) && (PRINT_VALUE) && (PRINT_VALUE_AT_OFFSET))
        {
            print_tag_address(ENTRY,current_offset,indent,"@",addrwidth);
            extraindent(SUBINDENT);
            printf("**** next IFD offset 0\n");
        }
    }

    if(ferror(inptr) || feof(inptr))
    {
        printf(" READ NEXT IFD OFFSET FAILED ");
        why(stdout);
        clearerr(inptr);    /* keep going...                          */
    }
    else
        current_offset = ftell(inptr);

    value_offset = current_offset;

    /* Make a second pass to print offset values.                     */
    if(use_second_pass)
    {
        if((PRINT_VALUE) && (PRINT_ADDRESS) && (PRINT_VALUE_AT_OFFSET))
        {
            print_tag_address(ENTRY,value_offset,indent,"@",addrwidth);
            extraindent(SUBINDENT);
            printf("============= VALUES, GPS IFD ============\n");
        }

        /* Second pass, to evaluate entries which are stored          */
        /* indirectly (the value requires more than 4 bytes).         */
        entry_offset = start_entry_offset;
        for(entry_num = 0; entry_num < num_entries; ++entry_num)
        {
            entry_ptr = read_entry(inptr,byteorder,entry_offset);
            if((entry_ptr->value_type == 0) || (entry_ptr->value_type > DOUBLE) ||
                ferror(inptr) || feof(inptr))
            {
                print_tag_address(ENTRY,entry_offset,indent,"@",addrwidth);
                printf(" INVALID (%lu)\n",entry_ptr->value);
                goto blewit;
            }
            current_offset = entry_offset = ftell(inptr);
            switch(entry_ptr->tag)
            {
                default:
                    if((PRINT_VALUE) && ((PRINT_VALUE_AT_OFFSET) && is_offset(entry_ptr)))
                    {
                        print_tag_address(ENTRY,fileoffset_base + entry_ptr->value,
                                            indent,"@",addrwidth);
                        print_tagid(entry_ptr,SUBINDENT,GPS_IFD);
                        value_offset =
                                    print_offset_value(inptr,byteorder,
                                        entry_ptr,fileoffset_base,
                                        GPS_IFD,indent,addrwidth,1);
                    }
                    break;
            }
            if(value_offset > max_value_offset)
                max_value_offset = value_offset;
        }
    }
    else if(current_offset > max_value_offset)
        max_value_offset = current_offset;

    if(max_offset && (gps_offset > max_offset))
    {
        print_tag_address(SECTION,max_value_offset - 1,indent,"*",addrwidth);
        max_value_offset = 0L;  /* let the caller know.               */
    }
    else
        print_tag_address(SECTION,max_value_offset - 1,indent,"-",addrwidth);
    if(PRINT_SECTION)
        printf("</GPS IFD>\n");
    STOPCOLOR;
    if(next_ifd_offset)
    {
        printred("#========= WARNING! CHAINING NEXT IFD FROM GPS IFD =========\n");
        value_offset = process_tiff_ifd(inptr,byteorder,next_ifd_offset,
                                fileoffset_base,max_offset,
                                GPS_IFD,-1,-1,indent,addrwidth);
        if(value_offset > max_value_offset)
            max_value_offset = value_offset;
    }
    return(max_value_offset);
blewit:
    clearerr(inptr);
    current_offset = ftell(inptr);
    print_tag_address(SECTION,current_offset - 1,indent,"-",addrwidth);
    if(PRINT_SECTION)
        printf("</GPS IFD>\n");
    STOPCOLOR;
    return(0L);
}

/* Process the marker segments between JPEG SOI and EOI               */
/* For the moment, that means just printing them (offset, id, length) */
/* and then moving to the next tag.                                   */

/* The first marker has been read, it's identity and offset are       */
/* passed in as arguments. No assumptions are made about the current  */
/* file pointer; the first read always begins at the passed           */
/* marker_offset.                                                     */

/* At least one image has now been observed "in the wild" which       */
/* includes APP segments within an included thumbnail (after edit by  */
/* PhotoShop), so this now attempts to indent such sub-segments to    */
/* indicate that structure.                                           */

unsigned long
process_jpeg_segments(FILE *inptr,unsigned long marker_offset,
                        unsigned short tag,unsigned long data_length,
                        int indent,int width)
{
    static unsigned long img_pixels = 0L;
    unsigned long max_offset = 0L;
    unsigned long start_of_jpeg_data;
    unsigned short seg_length,tmp;
    unsigned short img_height = 0;
    unsigned short img_width = 0;
    unsigned short bad_soi = 0;
    char *name;

    if(inptr)
    {
        start_of_jpeg_data = marker_offset;
        STARTCOLOR(JPEG_COLOR);
        while(tag != 0)
        {
            switch(tag)
            {
                case JPEG_SOS:  /* image data                         */
                    name = tagname(tag);
                    if(PRINT_SEGMENT)
                    {
                        print_tag_address(SEGMENT,marker_offset,indent + SUBINDENT,"@",width);
                        printf("<%s>",name);
                    }
                    seg_length = read_ushort(inptr,TIFF_MOTOROLA,marker_offset + 2);
                    if(ferror(inptr))
                    {
                        printf("ERROR reading length of SOS\n");
                        why(stdout);
                    }
                    else
                    {
                        if(PRINT_SEGMENT)
                        {
                            printf(" length %u",seg_length);
                            printf("  start of JPEG data, %lu pixels\n",img_pixels);
                        }
                        max_offset = ftell(inptr);
                    }
                    if(data_length > 0L)
                    {   
                        max_offset = ftell(inptr) + seg_length - 4;
                        marker_offset = start_of_jpeg_data + data_length - 2L;
                        tag = read_ushort(inptr,TIFF_MOTOROLA,marker_offset);
                        if(tag == JPEG_EOI)
                        {
                            max_offset = ftell(inptr);
                            continue;
                        }
                        else
                        {
                            if(fseek(inptr,0L,2) == 0)
                                max_offset = ftell(inptr);
                            else
                                max_offset = 0;
                        }
                    }
                    tag = 0;    /* break the loop                     */
                    break;
                case JPEG_COM:
                    name = tagname(tag);
                    if(PRINT_SEGMENT)
                    {
                        print_tag_address(SEGMENT,marker_offset,indent + SUBINDENT,"@",width);
                        printf("<%s>",name);
                    }
                    seg_length = read_ushort(inptr,TIFF_MOTOROLA,marker_offset + 2);

                    if(ferror(inptr) == 0)
                    {
                        if(PRINT_SEGMENT)
                        {
                            printf(" length %u: \"",seg_length);
                            (void)print_ascii(inptr,seg_length - 2,HERE);
                            printf("\"\n");
                        }
                        marker_offset += seg_length + 2;
                        /* next tag                                   */
                        tag = read_ushort(inptr,TIFF_MOTOROLA,marker_offset);
                        if(ferror(inptr))
                        {
                            tag = 0;
                            clearerr(inptr);
                        }
                        else max_offset = ftell(inptr);
                    }
                    else
                    {
                        tag = 0;
                        clearerr(inptr);
                    }
                    break;
                case JPEG_SOF_0:
                case JPEG_SOF_1:
                case JPEG_SOF_2:
                case JPEG_SOF_3:
                case JPEG_SOF_5:
                case JPEG_SOF_6:
                case JPEG_SOF_7:
                case JPEG_SOF_9:
                case JPEG_SOF_10:
                case JPEG_SOF_11:
                case JPEG_SOF_13:
                case JPEG_SOF_14:
                case JPEG_SOF_15:
                    name = tagname(tag);
                    if(PRINT_SEGMENT)
                    {
                        print_tag_address(SEGMENT,marker_offset,indent + SUBINDENT,"@",width);
                        printf("<%s>",name);
                    }
                    seg_length = read_ushort(inptr,TIFF_MOTOROLA,marker_offset + 2);
                    if(ferror(inptr) == 0)
                    {
                        if(PRINT_SEGMENT)
                            printf(" length %u",seg_length);
                        tmp = read_ubyte(inptr,HERE); /* precision    */
                        if(ferror(inptr) == 0)
                            img_height = read_ushort(inptr,TIFF_MOTOROLA,HERE);
                        if(ferror(inptr) == 0)
                            img_width = read_ushort(inptr,TIFF_MOTOROLA,HERE);
                        if(ferror(inptr) == 0)
                            img_pixels = img_height * img_width;
                        if(ferror(inptr) == 0)
                            tmp = read_ubyte(inptr,HERE); /* components */
                        if((ferror(inptr) == 0) && (PRINT_SEGMENT))
                            printf(", components=%d, height=%d, width=%d\n",
                                    (int)tmp,(int)img_height,(int)img_width);
                    }

                    if(ferror(inptr) == 0)
                    {
                        marker_offset += seg_length + 2;
                        /* next tag                                   */
                        tag = read_ushort(inptr,TIFF_MOTOROLA,marker_offset);
                        if(ferror(inptr))
                        {
                            tag = 0;
                            clearerr(inptr);
                        }
                        else max_offset = ftell(inptr);
                    }
                    else
                    {
                        tag = 0;
                        clearerr(inptr);
                    }
                    break;
                    
                case JPEG_BADSOI:   /* minolta                        */
                case JPEG_SOI:
                    name = tagname(tag);
                    if((PRINT_SECTION) || (PRINT_SEGMENT))
                    {
                        print_tag_address(SECTION|SEGMENT,marker_offset,indent,"@",width);
                        if(bad_soi)
                        {
                            printred("<JPEG_BADSOI>");
                            printf("(%#06x)\n",bad_soi);
                            bad_soi = 0;
                        }
                        else
                            printf("<%s>\n",name);
                    }
                    start_of_jpeg_data = marker_offset;
                    /* next tag                                       */
                    tag = read_ushort(inptr,TIFF_MOTOROLA,marker_offset + 2);
                    if(ferror(inptr) == 0)
                    {
                        marker_offset += 2;
                        max_offset = ftell(inptr);
                    }
                    else
                    {
                        tag = 0;
                        clearerr(inptr);
                    }
                    break;
                case JPEG_EOI:
                    name = tagname(tag);
                    max_offset = ftell(inptr);
                    if((PRINT_SECTION) || (PRINT_SEGMENT))
                    {
                        print_tag_address(SECTION|SEGMENT,marker_offset,indent,"@",width);
                        printf("<%s> JPEG length %lu\n",name,max_offset - start_of_jpeg_data);
                    }
                    tag = 0;  /* break the loop                       */
                    break;
                case JPEG_APP0: 
                    marker_offset = process_app0(inptr,marker_offset,tag,indent + SUBINDENT);
                    /* next tag                                       */
                    if(marker_offset > 0L)
                        tag = read_ushort(inptr,TIFF_MOTOROLA,marker_offset);
                    else
                        tag = 0;
                    if(ferror(inptr))
                    {
                        tag = 0;
                        clearerr(inptr);
                    }
                    else
                        max_offset = ftell(inptr);
                    break;
                case JPEG_APP1:
                    marker_offset = process_app1(inptr,marker_offset,tag,indent + SUBINDENT);
                    /* next tag                                       */
                    if(marker_offset > 0L)
                        tag = read_ushort(inptr,TIFF_MOTOROLA,marker_offset);
                    else
                        tag = 0;
                    if(ferror(inptr))
                        clearerr(inptr);
                    break;
                case JPEG_APP2:
                case JPEG_APP3:
                case JPEG_APP4:
                case JPEG_APP5:
                case JPEG_APP6:
                case JPEG_APP7:
                case JPEG_APP8:
                case JPEG_APP9:
                case JPEG_APP10:
                case JPEG_APP11:
                case JPEG_APP13:
                case JPEG_APP14:
                case JPEG_APP15:
                    marker_offset = process_appn(inptr,marker_offset,tag,indent + SUBINDENT);
                    /* next tag                                       */
                    if(marker_offset > 0L)
                        tag = read_ushort(inptr,TIFF_MOTOROLA,marker_offset);
                    else
                        tag = 0;
                    if(ferror(inptr))
                    {
                        tag = 0;
                        clearerr(inptr);
                    }
                    else max_offset = ftell(inptr);
                    break;
                case JPEG_APP12:
                    marker_offset = process_appn(inptr,marker_offset,tag,indent + SUBINDENT);
                    break;
                default:
                    /* Minolta sometimes writes garbage in the high   */
                    /* byte of SOI tags; assume that 0xd8 in the low  */
                    /* byte is a bad SOI.                             */
                    if(((tag & 0xff00) != 0xff00) && ((tag & 0x00ff) == 0xd8))
                    {
                        bad_soi = tag;
                        tag &= 0x00ff;
                        continue;
                    }
                    name = tagname(tag);
                    if(PRINT_SEGMENT)
                    {
                        print_tag_address(SEGMENT,marker_offset,indent + SUBINDENT,"@",width);
                        if((tag & 0xff00) != 0xff00)
                            printred(" INVALID JPEG TAG ");
                        printf("<%s>",name);
                    }
                    seg_length = read_ushort(inptr,TIFF_MOTOROLA,marker_offset + 2);
                    if(ferror(inptr) == 0)
                    {
                        if(PRINT_SEGMENT)
                            printf(" length %u\n",seg_length);
                        marker_offset = marker_offset + 2 + seg_length;
                        tag = read_ushort(inptr,TIFF_MOTOROLA,marker_offset);
                        if(ferror(inptr))
                        {
                            tag = 0;
                            clearerr(inptr);
                        }
                        else max_offset = ftell(inptr);
                    }
                    else
                    {
                        tag = 0;
                        clearerr(inptr);
                    }
                    break;
            }
        }
        STOPCOLOR;
    }
    return(max_offset);
}

/* JFIF format files                                                  */

unsigned long
process_app0(FILE *inptr,unsigned long app0_offset,unsigned short tag,
                int indent)
{
    unsigned long max_offset = 0L;
    unsigned long data_offset,dumplength;
    char *app_string,*name;
    unsigned app_length = 0L;
    unsigned short marker,xt,yt,tmp;

    if(inptr)
    {
        name = tagname(tag);
        app_length = read_ushort(inptr,TIFF_MOTOROLA,app0_offset + 2);
        print_tag_address(SECTION,app0_offset,indent,"@",ADDRWIDTH);
        data_offset = app0_offset + 4L;
        app_string = read_appstring(inptr,JPEG_APP0,data_offset);
        if(PRINT_SECTION)
        {
            printf("<%s> %#x length %u, ",name,tag,app_length);
            printf("\"%s\"",app_string);
        }
        if(PRINT_APPENTRY)
        {
            if((strncmp(app_string,"JFIF",4) == 0) && (app_string[4] == '\0'))
            {
                if(PRINT_SECTION)
                    (void)putchar('\n');
                data_offset += 5;
                print_tag_address(APP_ENTRY,data_offset,indent,"@",ADDRWIDTH);
                tmp = read_ushort(inptr,TIFF_MOTOROLA,data_offset);
                extraindent(SUBINDENT);
                printf("%-12.12s","Version");
                printf("= %d.%d\n",(int)((tmp >> 8) & 0xff),(int)(tmp & 0xff));
                data_offset = ftell(inptr);
                print_tag_address(APP_ENTRY,data_offset,indent,"@",ADDRWIDTH);
                tmp = read_ubyte(inptr,HERE);
                extraindent(SUBINDENT);
                printf("%-12.12s","Units");
                printf("= %s\n",tmp? tmp == 1 ? "dots/inch" : "dots/cm" : "aspect ratio");
                data_offset = ftell(inptr);
                print_tag_address(APP_ENTRY,data_offset,indent,"@",ADDRWIDTH);
                tmp = read_ushort(inptr,TIFF_MOTOROLA,HERE);
                extraindent(SUBINDENT);
                printf("%-12.12s","Xdensity");
                printf("= %u\n",tmp);
                data_offset = ftell(inptr);
                print_tag_address(APP_ENTRY,data_offset,indent,"@",ADDRWIDTH);
                tmp = read_ushort(inptr,TIFF_MOTOROLA,HERE);
                extraindent(SUBINDENT);
                printf("%-12.12s","Ydensity");
                printf("= %u\n",tmp);
                data_offset = ftell(inptr);
                print_tag_address(APP_ENTRY,data_offset,indent,"@",ADDRWIDTH);
                xt = read_ubyte(inptr,HERE);
                extraindent(SUBINDENT);
                printf("%-12.12s","XThumbnail");
                printf("= %d\n",(int)xt);
                data_offset = ftell(inptr);
                print_tag_address(APP_ENTRY,data_offset,indent,"@",ADDRWIDTH);
                yt = read_ubyte(inptr,HERE);
                extraindent(SUBINDENT);
                printf("%-12.12s","YThumbnail");
                printf("= %d\n",(int)yt);
                if(xt && yt)
                {
                    data_offset = ftell(inptr);
                    print_tag_address(APP_ENTRY,data_offset,indent,"@",ADDRWIDTH);
                    extraindent(SUBINDENT);
                    printf("%u RGB bytes (%u) pixels\n",3 * xt * yt,xt * yt);
                }
                max_offset = data_offset + 1 + (3 * xt * yt);
            }
            else if(strncmp(app_string,"JFXX",4) == 0)
            {
                if(PRINT_SECTION)
                    (void)putchar('\n');
                data_offset += 5;
                print_tag_address(APP_ENTRY,data_offset,indent,"@",ADDRWIDTH);
                tmp = read_ubyte(inptr,HERE);
                ++data_offset;
                extraindent(SUBINDENT);
                printf("extension code %#x",tmp);
                switch(tmp)
                {
                case 0x10:  printf(" - JPEG thumbnail\n");
                            marker = read_imagemarker(inptr,HERE); 
                            max_offset = process_jpeg_segments(inptr,
                                                data_offset,marker,
                                                app_length - 8,
                                                indent + SUBINDENT,
                                                ADDRWIDTH);
                            if(max_offset == 0)
                                max_offset = app0_offset + app_length + 2;
                            break;
                case 0x11:  printf(" - thumbnail stored using 1 byte/pixel\n"); 
                            max_offset = data_offset + 770 + (xt * yt) ;
                            break;
                case 0x13:  printf(" - thumbnail stored using 3 bytes/pixel\n");
                            max_offset = data_offset + (3 * xt * yt) ;
                            break;
                default:    printf(" - UNKNOWN JFXX extension\n");
                            max_offset = data_offset;
                            break;
                }
            }
            else
                goto appdump;

            /* If there's data unaccounted for, allow the data that isn't */
            /* accounted for to be dumped                                 */
            if(max_offset < (app0_offset + app_length + 2))
            {
                print_tag_address(SEGMENT,max_offset,indent,"@",ADDRWIDTH);
                extraindent(SUBINDENT);
                dumplength = app0_offset + app_length + 1 - max_offset;
                printf("---- End of data before end of APP0 (%lu bytes)\n",dumplength);
                /* dump at least the first couple of rows, just to see... */
                if(PRINT_APPNDUMP)
                {
                    if((Max_appdump == DUMPALL) || (Max_appdump > (app_length + 2)))
                        dumplength = app_length - 2;
                    else if(Max_appdump > 0L)
                        dumplength = Max_appdump;
                    else if(Max_undefined > 0L)
                        dumplength = Max_undefined;
                    else
                        dumplength = app_length - 2;
                }
                else
                    dumplength = dumplength > 48 ? 48 : dumplength;
                if(dumplength)
                {
                    hexdump(inptr,app0_offset + 4,app_length - 2,dumplength,16,
                                indent + SUBINDENT,SUBINDENT,ADDRWIDTH);
                    (void)putchar('\n');
                }
            }
        }
        else if(PRINT_SECTION)
            putchar('\n');
            
        /* close off the segment.                                     */
        if(Print_options & SECTION)
        {
            print_tag_address(SECTION,app0_offset + app_length + 1,indent,"@",ADDRWIDTH);
            printf("</%s>\n",name);
        }
    }

    return(app0_offset + app_length + 2);
appdump:
    /* Something went haywire...dump starting at the marker           */
    /* At the moment, this happens only for imaginative APP0 markers  */
    if(app_length > 0L)
    {
        if(PRINT_APPNDUMP)
        {
            if((Max_appdump == DUMPALL) || (Max_appdump > (app_length + 2)))
                dumplength = app_length + 2;
            else if(Max_appdump > 0L)
                dumplength = Max_appdump;
            else
                dumplength = Max_undefined;
            if(dumplength)
            {
                (void)putchar('\n');
                hexdump(inptr,app0_offset,app_length + 2,dumplength,16,
                            indent + SUBINDENT,SUBINDENT,ADDRWIDTH);
                (void)putchar('\n');
            }
        }
        else
        {
            if(PRINT_SECTION)
                printf(" - (not dumped: use -A)\n");
            max_offset = app0_offset + app_length + 2;
        }
        /* close off the segment; a start "tag" was printed           */
        print_tag_address(SECTION,app0_offset + app_length + 1,indent,"@",ADDRWIDTH);
        if(PRINT_SECTION)
            printf("</%s>\n",name);
    }
    return(max_offset);
}

/* Process (print) an APP1 (EXIF) section                             */

unsigned long
process_app1(FILE *inptr,unsigned long app1_offset,unsigned short tag,
                int indent)
{
    unsigned long max_offset = 0UL;
    unsigned long ifd_offset,fileoffset_base,current_offset;
    unsigned long dumplength = 0UL;
    unsigned short app_length = 0;
    unsigned short marker;
    char *app_string,*name;

    if(inptr)
    {
        /* display the APP1 header                                    */
        name = tagname(tag);    /* never null                         */
        print_tag_address(SECTION,app1_offset,indent,"@",ADDRWIDTH);
        if(PRINT_SECTION)
            printf("<%s> %#x ",name,tag);

        app_length = read_ushort(inptr,TIFF_MOTOROLA,HERE);
        if(ferror(inptr))
        {
            clearerr(inptr);
            goto appdump;
        }
        if(PRINT_SECTION)
            printf("length %u, ",app_length);

        app_string = read_appstring(inptr,JPEG_APP1,HERE);
        if(ferror(inptr))
        {
            clearerr(inptr);
            goto appdump;
        }

        if(app_string)
        {
            if(PRINT_SECTION)
                printf("\"%s\"",app_string);

            /* The next thing we see should be a TIFF header tag      */
            /* indicating byte order used in the IFD, followed by the */
            /* offset from the start of the header to the start of    */
            /* the IFD (number of entries).                           */

            fileoffset_base = ftell(inptr);
            /* If this is a valid Exif segment, skip the pad byte.    */
            if(strncmp(app_string,"Exif",4) == 0)
            {
                ++fileoffset_base;
                if(PRINT_SECTION)
                    putchar('\n');
                print_tag_address(SECTION,fileoffset_base,indent + SUBINDENT,"@",ADDRWIDTH);
                marker = read_imagemarker(inptr,fileoffset_base);
                if(marker && (print_header(marker,SECTION) == 0) &&
                    ((marker == TIFF_INTEL) || (marker == TIFF_MOTOROLA)))
                {
                    ifd_offset = read_ulong(inptr,marker,HERE);
                    if(PRINT_SECTION)
                        printf(" ifd offset = %lu (+ %lu = %#lx/%lu)\n",ifd_offset,
                                    fileoffset_base,fileoffset_base + ifd_offset,
                                    fileoffset_base + ifd_offset);
                    max_offset = process_tiff_ifd(inptr,marker,ifd_offset,
                                            fileoffset_base,0L,TIFF_IFD,
                                            0,-1,indent + SUBINDENT,ADDRWIDTH);

                    if(max_offset < app1_offset + app_length + 2)
                        max_offset = app1_offset + app_length +2;
                }
                else
                {
                    max_offset = app1_offset + app_length + 2;
                    if(PRINT_SECTION)
                    {
                        extraindent(indent + ADDRWIDTH);
                        printf(" INVALID TAG %u (%s) where TIFF header should be",
                                marker,tagname(marker));
#define PRINT_A_BIT 48
                        if(app_length > PRINT_A_BIT)
                            dumplength = PRINT_A_BIT;
                        else
                            dumplength = app_length;
                        goto appdump;
                    }
                }
            }
            else
            {
                max_offset = app1_offset + app_length + 2;
                if(PRINT_SECTION)
                {
                    printf(" - unknown format");
                    goto appdump;
                }
            }
        }
        else if(PRINT_SECTION)
            printf("  No APP1 header name string found\n");
        /* offset of last byte of segment                             */
        print_tag_address(SECTION,max_offset - 1,indent,"-",ADDRWIDTH);
        if(PRINT_SECTION)
            printf("</%s>\n",name);
    }
    return(max_offset);

appdump:
    if(app_length > 0L)
    {
        if((PRINT_APPNDUMP) || (dumplength > 0))
        {
            if((Max_appdump == DUMPALL) || (Max_appdump > (app_length + 2)))
                dumplength = app_length + 2;
            else if(Max_appdump > 0L)
                dumplength = Max_appdump;
            else if(dumplength == 0)
                dumplength = Max_undefined;
            if(dumplength)
            {
                (void)putchar('\n');
                hexdump(inptr,app1_offset,app_length + 2,dumplength,16,
                            indent + SUBINDENT,SUBINDENT,ADDRWIDTH);
                (void)putchar('\n');
            }
        }
        else
        {
            if(PRINT_SEGMENT)
                printf(" - (not dumped: use -A)\n");
            max_offset = app1_offset + app_length + 2;
        }
        /* close off the segment; a start "tag" was printed           */
        print_tag_address(SECTION,app1_offset + app_length + 1,indent,"-",ADDRWIDTH);
        if(PRINT_SECTION)
            printf("</%s>\n",name);
    }
    return(max_offset);
}
#undef PRINT_A_BIT


/* Unknown APPn markers; hex dump them and move on.                   */

unsigned long
process_appn(FILE *inptr,unsigned long appn_offset,unsigned short tag,
                int indent)
{
    char *app_string,*name;
    unsigned short app_length;
    unsigned long dumplength = 0L;

    if(inptr)
    {
        name = tagname(tag);    /* never null                         */
        app_length = read_ushort(inptr,TIFF_MOTOROLA,appn_offset + 2);
        print_tag_address(SECTION,appn_offset,indent,"@",ADDRWIDTH);
        if(PRINT_SECTION)
            printf("<%s> %#x length %u, ",name,tag,app_length);
        app_string = read_appstring(inptr,tag,HERE);
        if(PRINT_SECTION)
            printf("%s",app_string ? app_string : "NULL ID");

        if(PRINT_APPNDUMP)
        {
            if((Max_appdump == DUMPALL) || (Max_appdump > (app_length + 2)))
                dumplength = app_length + 2;
            else if(Max_appdump > 0L)
                dumplength = Max_appdump;
            else
                dumplength = Max_undefined;
            if(dumplength)
            {
                (void)putchar('\n');
                hexdump(inptr,appn_offset,app_length + 2,dumplength,16,
                            indent,SUBINDENT,ADDRWIDTH);
                (void)putchar('\n');
            }
        }
        else if(PRINT_SECTION)
            printf(" - unknown format (not dumped: use -A)\n");

        print_tag_address(SECTION,appn_offset + app_length + 1,indent,"-",ADDRWIDTH);
        if(PRINT_SECTION)
            printf("</%s>\n",name);
    }
    return(appn_offset + app_length +2);
}

/* Print the string data from an APP12 segment, if hexdump hasn't     */
/* been asked. The "[fileinfo]" stuff is unknown format, but this     */
/* part should be useful.                                             */

unsigned long
process_app12(FILE *inptr,unsigned long appn_offset,unsigned short tag,
                int indent)
{
    char *app_string,*name;
    unsigned short app_length;
    unsigned long dumplength = 0UL;
    unsigned long offset = 0UL;
    unsigned long end_offset = 0UL;
    unsigned long size = 0UL;
    int nextch;

    if(inptr)
    {
        name = tagname(tag);    /* never null                         */
        app_length = read_ushort(inptr,TIFF_MOTOROLA,appn_offset + 2);
        print_tag_address(SECTION,appn_offset,indent,"@",ADDRWIDTH);
        if(PRINT_SECTION)
            printf("<%s> %#x length %u, ",name,tag,app_length);
        app_string = read_appstring(inptr,tag,HERE);
        if(PRINT_SECTION)
            printf("%s",app_string ? app_string : "NULL ID");

        if(PRINT_APPNDUMP)
        {
            if((Max_appdump == DUMPALL) || (Max_appdump > (app_length + 2)))
                dumplength = app_length + 2;
            else if(Max_appdump > 0L)
                dumplength = Max_appdump;
            else
                dumplength = Max_undefined;
            if(dumplength)
            {
                if(PRINT_SECTION)
                    putchar('\n');
                hexdump(inptr,appn_offset,app_length + 2,dumplength,16,
                            indent + SUBINDENT,SUBINDENT,ADDRWIDTH);
                (void)putchar('\n');
            }
        }
        else if(PRINT_APPENTRY)
        {
            end_offset = appn_offset + app_length + 1;
            offset = ftell(inptr);
            nextch = skip_past_newline(inptr,end_offset);
            offset = ftell(inptr);
            if(PRINT_SECTION)
                putchar('\n');
            nextch = putword(inptr,nextch,end_offset,indent);
            putchar('\n');
            offset = ftell(inptr);
            while(nextch && (offset < end_offset))
            {
                nextch = putword(inptr,nextch,end_offset,indent);
                if(nextch)
                    putchar('\n');
                offset = ftell(inptr);
            }
            while(offset < end_offset)
            {
                nextch = skip_to_bracket(inptr,end_offset);
                size = ftell(inptr) - offset;
                if(nextch == '[')
                    --size;
                if(size > 1)
                {
                    printf(" length %lu",size);
                    if(Max_undefined)
                    {
                        putchar('\n');
                        hexdump(inptr,offset,size, Max_undefined,16,
                                    indent + SUBINDENT,
                                    SUBINDENT,ADDRWIDTH);
                        putchar('\n');
                        nextch = fgetc(inptr);
                        if(ftell(inptr) >= end_offset)
                            break;
                    }
                    else
                        printred(" (not dumped; use -U)\n");
                }
                else
                    putchar('\n');
                if(nextch)
                    nextch = putword(inptr,nextch,end_offset,indent);
                offset = ftell(inptr);
            }
        }

        print_tag_address(SECTION,appn_offset + app_length + 1,indent,"-",ADDRWIDTH);
        if(PRINT_SECTION)
            printf("</%s>\n",name);
    }
    return(appn_offset + app_length +2);
}

/* Read and discard until the next newline in the input stream, or    */
/* until 'max_offset' is reached. Return the last character read.     */

int
skip_past_newline(FILE *inptr,unsigned long max_offset)
{
    while(((unsigned long)ftell(inptr) < max_offset) && (fgetc(inptr) != '\n'))
        continue;
    return(fgetc(inptr));
}

/* Read and discard until the next '[' in the input stream, or until  */
/* 'max_offset' is reached. Return the last character read.           */

int
skip_to_bracket(FILE *inptr,unsigned long max_offset)
{
    int ch = EOF;

    while(((unsigned long)ftell(inptr) < max_offset) && ((ch = fgetc(inptr)) != '['))
        continue;
    if(ftell(inptr) >= max_offset)
        ch = 0;
    return(ch);
}

/* Print a word from an APP12 segment. The first character of the     */
/* word is passed in by the caller; if it's a left square bracket, it */
/* is assumed to start a "section" and is indented a little less.     */
/* This can print a lone open-bracket, but what the heck.             */

/* CRs are stripped out and nulls end a word. Newlines are stripped   */
/* as well, but cause the next character to be read and passed back   */
/* to the caller, to be used as the start of a new word. Otherwise    */
/* the last character read is returned.                               */

/* The routine is not allowed to read past 'max_offset', which should */
/* be the end of the segment.                                         */

int
putword(FILE *inptr,int firstch,unsigned long max_offset,int indent)
{
    int eol = 0;
    int ch = firstch;

    print_tag_address(APP_ENTRY,ftell(inptr),indent + SUBINDENT,"@",ADDRWIDTH);
    if(firstch != '[')
        extraindent(SUBINDENT);
    putchar(firstch);
    while(((unsigned long)ftell(inptr) < max_offset) && !eol)
    {
        ch = fgetc(inptr);
        switch(ch)
        {
        case '\r':
            break;
        case 0:
            ++eol;
            break;
        case '\n':
            ++eol;
            ch = fgetc(inptr);
            break;
        case EOF:
            ++eol;
            break;
        default:
            if(isprint(ch))
                putchar(ch);
            break;
        }
    }
    return(ch);
}



/* Determine if the values for an IFD entry are recorded at an        */
/* offset. Returns 0 if the value is recorded directly in the entry,  */
/* 1 if the value is offset.                                          */

int
is_offset(struct ifd_entry *entry_ptr)
{
    if((value_type_size(entry_ptr->value_type) * entry_ptr->count) > 4)
        return(1);
    return(0);
}
